﻿/* This Software is Licenced under MIT Licence (http://en.wikipedia.org/wiki/MIT_License)

Copyright (c) 2012 Andrea Cappelletti, Milan, Italy aka CleverWalrus)

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge, 
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons 
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
DEALINGS IN THE SOFTWARE.*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

using CleverBobCat.Core;

namespace CleverBobCat.EngineController
{
    public class ModuleCleverRCSModule : PartModule
    {
        [KSPField]
        public bool debug = false;

        [KSPField]
        public string engineGroup = "default";

        [KSPField]
        public bool engineActive = false;

        [KSPField]
        public float engineThrottle = 0F;

        // maxThrust of engine can't be == to minThrust
        public float minValue = 0.00000001F;

        // RCS Modules don't have a single thrust but up to 6
        // thrustVectors we save them all with original state
        private float[] thrusterPowers;
        private float[] origThrusterPowers;
        private float fuelConsumption;
        private float origFuelConsumption;

        // We use this for avoiding running the script unless a
        // ModuleCleverEngineController has taken control of engine
        public bool controlled = false;

        /// <summary>
        /// Constructor style setup.
        /// Per-frame update
        /// Called even if part is not active
        /// </summary>
        void Update()
        {
        }

        /// <summary>
        /// Constructor style setup.
        /// Per-physx-frame update
        /// Called even if part is not active
        /// </summary>
        void FixedUpdate()
        {
            // checking controlled avoid to run the script if the vessel
            // is not the one that was activated in this scene
            if (controlled)
            {
                // We must update the FX even for non-active vessels
                // rcs can be used even if not "active" so we use
                // FixedUpdate that run even on non active parts
                UpdateEngineFX();
            }
        }

        /// <summary>
        /// Constructor style setup.
        /// Called in the Part\'s Awake method. 
        /// The model may not be built by this point.
        /// </summary>
        public override void OnAwake()
        {
        }

        /// <summary>
        /// Called during the Part startup.
        /// StartState gives flag values of initial state
        /// </summary>
        public override void OnStart(StartState state)
        {
            Logger.Debug(debug, "CleverRCSModule.OnStart");
            RCSModule rcs = (RCSModule)part;
            // We take the "max" thrust of each rcs thrustVector from the vector themselfs
            origThrusterPowers = new float[6];
            origThrusterPowers[0] = rcs.thrustVector0.w;
            origThrusterPowers[1] = rcs.thrustVector1.w;
            origThrusterPowers[2] = rcs.thrustVector2.w;
            origThrusterPowers[3] = rcs.thrustVector3.w;
            origThrusterPowers[4] = rcs.thrustVector4.w;
            origThrusterPowers[5] = rcs.thrustVector5.w;
            thrusterPowers = new float[6];
            thrusterPowers[0] = rcs.thrustVector0.w;
            thrusterPowers[1] = rcs.thrustVector1.w;
            thrusterPowers[2] = rcs.thrustVector2.w;
            thrusterPowers[3] = rcs.thrustVector3.w;
            thrusterPowers[4] = rcs.thrustVector4.w;
            thrusterPowers[5] = rcs.thrustVector5.w;
            // We take the original fuel consumption since RCS burn fuel even if thrust is 0
            fuelConsumption = rcs.fuelConsumption;
            origFuelConsumption = rcs.fuelConsumption;
            for (int i = 0; i < 6; i++)
            {
                Logger.Debug(debug, "thrusterPowers[" + i + "]", thrusterPowers[i].ToString());
            }
            Logger.Debug(debug, "fuelConsumption", fuelConsumption.ToString());
        }

        /// <summary>
        /// Per-frame update
        /// Called ONLY when Part is ACTIVE!
        /// </summary>
        public override void OnUpdate()
        {
        }

        /// <summary>
        /// Per-physx-frame update
        /// Called ONLY when Part is ACTIVE!
        /// </summary>
        public override void OnFixedUpdate()
        {
        }

        /// <summary>
        /// Called when PartModule is asked to save its values.
        /// Can save additional data here.
        /// </summary>
        /// <param name='node'>The node to save in to</param>
        public override void OnSave(ConfigNode node)
        {
        }

        /// <summary>
        /// Called when PartModule is asked to load its values.
        /// Can load additional data here.
        /// </summary>
        /// <param name='node'>The node to load from</param>
        public override void OnLoad(ConfigNode node)
        {
        }

        /// <summary>
        /// Called from ModuleCleverEngineController for changing
        /// the thrust and fuel consumption of the engine using toggle
        /// and slider setting this does not change the value
        /// for the engine but only the local vars.
        /// for rcs we must handle the fuel consumption too unlike other engines
        /// </summary>
        [KSPEvent]
        public void UpdateEngineGroup(BaseEventData data)
        {
            controlled = true;
            CleverEngineGroup engineGroupData = (CleverEngineGroup)data.Get("engineGroup");
            if (engineGroupData.name == engineGroup)
            {
                Logger.Debug(debug, "UpdateEngineGroup", "CleverRCSModule");
                if (engineGroupData.active)
                {
                    engineActive = true;
                    engineThrottle = engineGroupData.throttle;
                    for (int i = 0; i < 6; i++)
                    {
                        Logger.Debug(debug, "origThrusterPowers[" + i + "]", origThrusterPowers[i].ToString());
                        thrusterPowers[i] = (origThrusterPowers[i] * engineThrottle) / 100;
                    }
                    fuelConsumption = (origFuelConsumption * engineThrottle) / 100;
                    LogStatus();
                }
                else
                {
                    engineActive = false;
                    engineThrottle = engineGroupData.throttle;
                    for (int i = 0; i < 6; i++)
                    {
                        thrusterPowers[i] = 0;
                    }
                    fuelConsumption = 0;
                    LogStatus();
                }
                for (int i = 0; i < 6; i++)
                {
                    Logger.Debug(debug, "thrusterPowers[" + i + "]", thrusterPowers[i].ToString());
                }
                Logger.Debug(debug, "fuelConsumption", fuelConsumption.ToString());

                // Update the actual engine values
                UpdateEngineThrust();
            }
        }

        /// <summary>
        /// Called for changing the real parameters of the engine.
        /// we set the thrusterPower for all 6 possible rcs thrustVectors
        /// </summary>
        private void UpdateEngineThrust()
        {
            RCSModule rcs = (RCSModule)part;
            for (int i = 0; i < 6; i++)
            {
                rcs.thrusterPowers[i] = thrusterPowers[i];
                
            }
            rcs.fuelConsumption = fuelConsumption;
        }

        /// <summary>
        /// Called for disabling the FX when engine is disabled
        /// the "game" throttle otherwise set the FX plume
        /// </summary>
        private void UpdateEngineFX()
        {
            if (!engineActive || thrusterPowers[0] == 0) part.findFxGroup("rcsGroup0").setActive(false);
            if (!engineActive || thrusterPowers[1] == 0) part.findFxGroup("rcsGroup1").setActive(false);
            if (!engineActive || thrusterPowers[2] == 0) part.findFxGroup("rcsGroup2").setActive(false);
            if (!engineActive || thrusterPowers[3] == 0) part.findFxGroup("rcsGroup3").setActive(false);
            if (!engineActive || thrusterPowers[4] == 0) part.findFxGroup("rcsGroup4").setActive(false);
            if (!engineActive || thrusterPowers[5] == 0) part.findFxGroup("rcsGroup5").setActive(false);
        }

        private void LogStatus()
        {
            Logger.Debug(debug, "EngineGroup:[" + engineGroup + "] Active:[" + engineActive + "] Throttle:[" + engineThrottle.ToString() + "]");
        }
    }
}
