﻿/* Derived from CART.DLL, Cart Plugin for KSP.
Original Author: Tony Chernogovsky, SPb, Russia.
Previous Maintainer: Callum King-Underwood, England.
Current Maintainer: Andrea Cappelletti, Milan, Italy aka CleverWalrus.
License: CC-BY-SA, Attribution & Share-alike Creative Common Licence. 
Feel free to modify as long as:
a) original author is mentioned, and 
b) you distribute your code under the same license.*/
  
/* See http://en.wikipedia.org/wiki/PID_controller#Manual_tuning for PID setup strategy.
Note 'integrationBuffer' parameter. It determines for how long PID will remember its 
'history'. Set to 0 to remember _each_ point registered (which is a standard PID behaviour).
'Clamp' parameter simply limits the output value.*/

using UnityEngine;
namespace CleverBobCat.LegacyCart
{
    public class PidController
    {
        public /* private */ float mKp, mKd, mKi;
        private float mOldVal, mOldTime, mOldD;
        private float mClamp;
        private float[] mBuffer = null;
        private int mPtr;
        private float mSum;
        private float mValue;

        public PidController(float Kp, float Ki, float Kd,
                              int integrationBuffer, float clamp)
        {
            mKp = Kp;
            mKi = Ki;
            mKd = Kd;
            mClamp = clamp;
            if (integrationBuffer >= 1)
                mBuffer = new float[integrationBuffer];
            Reset();
        }

        public void Reset()
        {
            mSum = 0;
            mOldTime = -1;
            mOldD = 0;
            if (mBuffer != null)
                for (int i = 0; i < mBuffer.Length; i++)
                    mBuffer[i] = 0;
            mPtr = 0;
        }

        public float Control(float v)
        {
            if (Time.fixedTime > mOldTime)
            {
                if (mOldTime >= 0)
                {
                    mOldD = (v - mOldVal) / (Time.fixedTime - mOldTime);

                    float i = v / (Time.fixedTime - mOldTime);
                    if (mBuffer != null)
                    {
                        mSum -= mBuffer[mPtr];
                        mBuffer[mPtr] = i;
                        mPtr++;
                        if (mPtr >= mBuffer.Length)
                            mPtr = 0;
                    }
                    mSum += i;
                }

                mOldTime = Time.fixedTime;
                mOldVal = value;
            }

            mValue = mKp * v + mKi * mSum + mKd * mOldD;

            if (mClamp > 0)
            {
                if (mValue > mClamp)
                    mValue = mClamp;
                if (mValue < -mClamp)
                    mValue = -mClamp;
            }

            return mValue;
        }

        public float value
        {
            get { return mValue; }
        }

        public static implicit operator float(PidController v)
        {
            return v.mValue;
        }
    }
}